function table.pack(...)
  return { n = select("#", ...), ... }
end

-- Debug
local in_logout = false
local in_updateFac = false
local in_loaded = false

local color = {}
color.red = "FFFF0000"
color.green = "FF00FF00"
color.blue = "FF0000FF"
color.brown = "FFFFFF00"
color.purple = "FFFF00FF"
color.orange = "FFe26a24"
color.yellow = "FFfff20c"
color.gray = "FF595046"

local TAB = "      "

local function debugPrint(...)
	if false then
		local str = ""
		local args = table.pack(...)
		for i=1,args.n do
			local v = args[i]
			local t = type(v)
			if t == "string" or t == "number" then
				str = str .. v .. " "
			elseif t == "boolean" then
				if v then str = str .. "true" else str = str .. "false" end
			else
				str = str .. t
			end
		end
		ChatFrame5:AddMessage(str .. "\n")
	end
end

local function colorString(str, color)
	return "|c" .. color .. str .. "|r"
end

local PARAGON_REP_DATA = {
	[2170] = {name = "Argussian Reach", rewardQuestID = 48976, threshold = 10000, nicknames = {"argussian", "argus", "reach", "ar"}, color = "ffa731b7"},
	[2045] = {name = "Armies of Legionfall", rewardQuestID = 46777, threshold = 10000, nicknames = {"armies", "legionfall", "aol"}, color = "ff4e6d32"},
	[2165] = {name = "Army of the Light", rewardQuestID = 48977, threshold = 10000, nicknames = {"army", "light", "aotl"}, color = "ffd3c310"},
	[1900] = {name = "Court of Farondis", rewardQuestID = 46745, threshold = 10000, nicknames = {"court", "farondis", "azsuna", "aszuna", "cof"}, color = "ff2cd4e0"},
	[1883] = {name = "Dreamweavers", rewardQuestID = 46747, threshold = 10000, nicknames = {"dreamweavers", "dream", "valsharah", "val'sharah", "dw"}, color = "ff3de274"},
	[1828] = {name = "Highmountain Tribes", rewardQuestID = 46743, threshold = 10000, nicknames = {"highmountain", "tribes", "cows", "hm"}, color = "ff7c592b"},
	[1859] = {name = "The Nightfallen", rewardQuestID = 46748, threshold = 10000, nicknames = {"nightfallen", "nightborne", "nightborn", "suramar", "nf", "nb"}, color = "ff2e6f84"},
	[1894] = {name = "The Wardens", rewardQuestID = 46749, threshold = 10000, nicknames = {"wardens", "w"}, color = "ff54656b"},
	[1948] = {name = "Valarjar", rewardQuestID = 46746, threshold = 10000, nicknames = {"valarjar", "stormheim", "vj"}, color = "fff29e37"},
}

local old_UnitFullName = UnitFullName
local function UnitFullName()
	local name, _ = old_UnitFullName("player") -- 2nd param "realm" is typically nil for some time after the addon isloaded, so GetRealmName must be used instead
	local realm = GetRealmName()
	return name, realm
end

--Usable as a key to get data on the player from saved data.
local function getPlayerStorageKey()
	local name, _ = UnitFullName("player") -- 2nd param "realm" is typically nil for some time after the addon isloaded, so GetRealmName must be used instead
	local realm = GetRealmName()
	return name .. "-" .. realm
end

local function createBaseCharacterData(name)
	local general = {}

	local localizedClass, englishClass, classIndex = UnitClass("player")
	general.localizedClass = localizedClass
	general.englishClass = englishClass
	general.classIndex = classIndex

	local name, realm = UnitFullName("player")
	general.name = name
	general.name_server = name .. "-" .. realm
	--general.coloredName = function(general) return colorString(general.name, RAID_CLASS_COLORS[general.englishClass].colorStr) end

	return {general = general, paragonRep = {}}
end

local function setUpCharacterDataIfNotPresent()
	local storageKey = getPlayerStorageKey()
	if not ParagonRepTrackerData.characterData[storageKey] then
		ParagonRepTrackerData.characterData[storageKey] = createBaseCharacterData()
	end
end

-- updates paragon rep information in ParagonRepTrackerData.characterData.
local function updateParagonRep()
	local paragonRep = ParagonRepTrackerData.characterData[getPlayerStorageKey()].paragonRep
	for i in pairs(PARAGON_REP_DATA) do
		local totalRep, _, _, hasRewardPending, tooLowLevelForParagon = C_Reputation.GetFactionParagonInfo(i)

		if paragonRep[i] then
			if paragonRep[i].hasParagon and not totalRep then
				if not ParagonRepTrackerData.debug then ParagonRepTrackerData.debug = {} end
				local text = "unParagoned for " .. ParagonRepTrackerData.characterData[getPlayerStorageKey()].general.name .. " in event " .. toString(in_logout) .. " " .. toString(in_loading) .. " " .. toString(in_updateFac)
				debugPrint(colorString(text, color.red))
				ParagonRepTrackerData.debug[#ParagonRepTrackerData.debug + 1] = text
			end
		end

		if totalRep then
			paragonRep[i] = {hasParagon = true, totalRep = totalRep, hasRewardPending = hasRewardPending, tooLowLevelForParagon = tooLowLevelForParagon}
		else
			paragonRep[i] = {hasParagon = false}
		end
	end
end

local function printUsage()
	print("usage of /parrep")
	print("`/parrep` - prints this message.")
	print("`/parrep max` - prints which character has the highest paragon for each faction.")
	print("`/parrep <character name>` - prints paragon information for that character. You can also do `/parrep me` for your own paragon reputation, or `/parrep all` to see data on all characters on your account.")
	print("`/parrep <faction>` - prints paragon rep information for all your characters for a particular faction. To find what arguments are valid for `<faction>`, see output of `/parrep factionargs`.")
end

local function getNamesAndFactions(txt)
	local names = {}
	local factions = {}
	for arg in txt:gmatch("%S+") do
		local lowercaseArg = arg:lower()

		if lowercaseArg == "me" then
			names[getPlayerStorageKey()] = true
		else 

			local foundName = false
			for j in pairs(ParagonRepTrackerData.characterData) do
				local fullName = j:lower()
				local shortName = fullName:sub(0, fullName:find("-") - 1)
				if lowercaseArg == fullName or lowercaseArg == shortName then
					names[j] = true
					foundName = true
				end
			end

			if not foundName then 
				if (lowercaseArg == "all") then
					names = {}
					for name in pairs(ParagonRepTrackerData.characterData) do
						names[name] = true
					end
				else
					local foundFaction = false
					for factionID,data in pairs(PARAGON_REP_DATA) do
						for i in ipairs(data.nicknames) do
							if lowercaseArg == data.nicknames[i] then
								factions[factionID] = true
								foundFaction = true
								break -- not necessary since there should not be duplicate nicknames, but doesn't hurt.
							end
						end
					end
					if not foundFaction then
						print(colorString("Invalid argument " .. arg, color.red))
						printUsage()
						return
					end
				end
			end
		end
	end
	return names, factions
end

local function makeRepString(factionID, repData, charName)
	local result = ""
	local repName
	if charName then
		repName = colorString(charName, RAID_CLASS_COLORS[ParagonRepTrackerData.characterData[charName].general.englishClass].colorStr)
	else
		repName = colorString(PARAGON_REP_DATA[factionID].name, PARAGON_REP_DATA[factionID].color)
	end
	result = result .. repName
	if repData.hasParagon == nil then
		result = result .. colorString(" This addon has seen this character before, but not this reputation.", color.orange) -- TODO remove this condition? It shouldn't happen.
	elseif not repData.hasParagon then
		local text = colorString("NO PARAGON", color.gray)
		result = result .. " " .. text
	else
		local threshold = PARAGON_REP_DATA[factionID].threshold
		local paragonProgress = mod(repData.totalRep, threshold)
		result = result .. " " .. paragonProgress --.. "/" .. threshold
		if repData.hasRewardPending then
			result = result .. " " .. colorString("REWARD READY", color.green)
		end
		if repData.tooLowLevelForParagon then
			result = result .. " " .. colorString("(TOO LOW LEVEL FOR PARAGON)", color.red)
		end
	end
	return result
end

local function printByFaction(names, factions)
	for factionID in pairs(factions) do

		local charReps = {}

		local NO_PARAGON = colorString("NO PARAGON: ", color.gray)
		for name in pairs(names) do
			local cData = ParagonRepTrackerData.characterData[name]
			local repData = cData.paragonRep[factionID]
			if repData then
				local repString = makeRepString(factionID, repData, cData.general.name_server)
				local amt
				if not repData.hasParagon then
					NO_PARAGON = NO_PARAGON .. colorString(ParagonRepTrackerData.characterData[name].general.name, RAID_CLASS_COLORS[ParagonRepTrackerData.characterData[name].general.englishClass].colorStr) .. ", "
				else
					local threshold = PARAGON_REP_DATA[factionID].threshold
					local paragonProgress = mod(repData.totalRep, threshold)
					charReps[#charReps + 1] = {str = repString, amt = paragonProgress}
				end
			end
		end

		table.sort(charReps, function(l, r) return l.amt < r.amt end)
		print(colorString("##### " .. PARAGON_REP_DATA[factionID].name .. " #####", PARAGON_REP_DATA[factionID].color))
		print(TAB .. NO_PARAGON:sub(0, NO_PARAGON:len() - 2))
		for _,b in pairs(charReps) do
			print(TAB .. b.str)
		end
	end
end

local function printByCharacter(names, factions)
	for name in pairs(names) do

		local cData = ParagonRepTrackerData.characterData[name]

		local allReps = cData.paragonRep
		local coloredName = colorString("##### " .. name .. " #####", RAID_CLASS_COLORS[cData.general.englishClass].colorStr)
		print(coloredName)

		for factionID in pairs(factions) do
			local repData = allReps[factionID]
			if repData then
				local result = makeRepString(factionID, repData)
				print(TAB .. result)
			end
		end
	end
end

local function getHighestRep(factionID)
	local highest = -1
	local highestData = nil
	local highestName = nil
	for name,cData in pairs(ParagonRepTrackerData.characterData) do
		local repData = cData.paragonRep[factionID]
		if repData and repData.hasParagon then
			local paragonProgress = mod(repData.totalRep, PARAGON_REP_DATA[factionID].threshold)
			if paragonProgress > highest then
				highest = paragonProgress
				highestData = repData
				highestName = name
			end
		elseif repData and not repData.hasParagon then
			if highest == -1 then
				highestData = repData
				highestName = name
			end
		end
	end
	return highestName, highestData
end

local function printMaxReps()
	print(colorString("=========== MAX REP ==========", color.orange))

	for factionID,data in pairs(PARAGON_REP_DATA) do
		local name, repData = getHighestRep(factionID)
		local result = colorString(PARAGON_REP_DATA[factionID].name .. ":", PARAGON_REP_DATA[factionID].color)
		if name == nil then
			result = result .. " " .. colorString("no data for this faction", color.gray)
		elseif repData.hasParagon then
			result = result .. " " .. makeRepString(factionID, repData, name)
		else
			result = result .. " " .. colorString("NO PARAGON", color.gray)
		end
		print(TAB .. result)
	end
end

local function printFactionArgs()
	print("possible arguments for <faction>:")
	for _,data in pairs(PARAGON_REP_DATA) do
		local result = colorString(data.name, data.color) .. ": "
		for i,nickname in ipairs(data.nicknames) do
			result = result .. nickname .. ", "
		end
		print(result:sub(0, result:len() - 2))
	end
end

SLASH_PARREP1 = "/parrep"
SlashCmdList["PARREP"] = function(txt, editbox)
	local numArgs = 0
	for arg in txt:gmatch("%S+") do
		debugPrint("arg:", '"' .. arg:lower() .. '"', '"' .. "max" .. '"', arg:lower() == "max")
		numArgs = numArgs + 1
		if arg:lower() == "factionargs" then
			printFactionArgs()
			return
		end
		if arg:lower() == "max" then
			printMaxReps()
			return
		end
	end
	if numArgs == 0 then
		printUsage()
		return
	end

	local names, factions = getNamesAndFactions(txt)
	if names == nil then return end

	local result = "names:" for name in pairs(names) do result = result .. name .. ", " end debugPrint(result)
	result = "factions:" for faction in pairs(factions) do result = result .. faction .. ", " end debugPrint(result)

	--print(colorString("=========== PARAGON REPUTATION ==========", color.green))

	if next(factions) == nil then
		factions = PARAGON_REP_DATA
	else
		if next(names) == nil then
			names = {}
			for name in pairs(ParagonRepTrackerData.characterData) do
				names[name] = true
			end
			printByFaction(names, factions)
			return
		end
	end
	if next(names) == nil then names = {[getPlayerStorageKey()] = true} end

	printByCharacter(names, factions)
end

-- TODO /parrep max.
-- TODO tooltips.

-- Registers event handlers to an anonymous frame.
local function registerEventHandlers(events)
	local frame = CreateFrame("Frame")
	frame:SetScript("OnEvent", function(self, event, ...)
	 events[event](self, ...); -- call one of the functions above
	end);
	for k, v in pairs(events) do
	 frame:RegisterEvent(k); -- Register all events for which handlers have been defined
	end
end

-- TODO why do local functions have to be declared prior? I think I know but I should at least figure out how to handle this (how to order functions, forwards declarations, etc.)
--   And why it seems to work differently for ":" than ".".
-- TODO wrong data being saved (Cannot replicate anymore).
--   I wonder if I'm updating too early/too late?

local events = {}
function events:ADDON_LOADED(...)
	if select(1, ...) ~= "ParagonRepTracker" then return end
	debugPrint("ADDON_LOADED")
	if ParagonRepTrackerData then
		debugPrint("found saved data")
	else
		debugPrint("no saved data found")
	end
	debugPrint("PRTD:", ParagonRepTrackerData)

	if not ParagonRepTrackerData then
		ParagonRepTrackerData = {
			characterData = {}
		}
	end

	local a = ParagonRepTrackerData.debug
	if a then
		for i=1,#a do
			debugPrint("a: " .. colorString(a[i], color.red))
		end
	end

	in_loaded = true
	setUpCharacterDataIfNotPresent()
	updateParagonRep()
	in_loaded = false
end
function events:UPDATE_FACTION(...)
   --print("UPDATE_FACTION")
	in_updateFac = true
	updateParagonRep()
	in_updateFac = false
end
function events:PLAYER_LOGOUT(...)
	in_logout = true
	debugPrint("PLAYER_LOGOUT")
	updateParagonRep() -- TODO comment out.
	in_logout = false
end
registerEventHandlers(events)
